"""Unit Tests for inferences module"""


import pytest
import numpy as np
import pandas as pd
from pandas.testing import assert_series_equal
from statsmodels.tsa.arima_process import ArmaProcess

import causalimpact
from unittest.mock import Mock

from causalimpact.model import ModelResults

compile_inferences = causalimpact.inferences.compile_inferences
np.random.seed(1)


@pytest.fixture
def data():
    ar = np.r_[1, 0.9]
    ma = np.array([1])
    arma_process = ArmaProcess(ar, ma)

    X = 1 + arma_process.generate_sample(nsample=100)
    X = X.reshape(-1, 1)
    y = 1.2 * X + np.random.normal(size=(100, 1))
    data = np.concatenate((y, X), axis=1)
    data = pd.DataFrame(data)
    return data


@pytest.fixture
def pre_period():
    return [0, 49]


@pytest.fixture
def post_period():
    return [50, 100]


@pytest.fixture
def trained_model(post_period):
    trained_model = Mock(spec=ModelResults)
    trained_model.model_nobs = 100
    # the mocked methods on these two classses
    # only work because pre-period and
    # post_period are both 50 length

    class PredictionResultsMock:
        def __init__(self, n):
            self.predicted_mean = np.ones(50)

        def conf_int(self, alpha=None):
            return np.ones(100).reshape(50, 2)

    class ForecastResultsMock:
        def __init__(self):
            self.predicted_mean = np.ones(50)

        def conf_int(self, alpha=None):
            return np.ones(100).reshape(50, 2)

    def get_prediction_mock(start=None, end=None):
        if start is not None or end is not None:
            return PredictionResultsMock(n=50)
        else:
            return PredictionResultsMock(n=100)

    def get_forecast_mock(df_post, alpha=None):
        return ForecastResultsMock()

    trained_model.get_prediction = get_prediction_mock
    trained_model.get_forecast = get_forecast_mock
    return trained_model


def test_compile_inferences_w_data(data, pre_period, post_period, trained_model):

    df_pre = data.loc[pre_period[0] : pre_period[1], :]
    df_post = data.loc[post_period[0] : post_period[1], :]

    post_period_response = None
    alpha = 0.05
    orig_std_params = (0.0, 1.0)

    estimation = "MLE"

    inferences = compile_inferences(
        trained_model,
        data,
        df_pre,
        df_post,
        post_period_response,
        alpha,
        orig_std_params,
        estimation,
    )

    expected_response = pd.Series(data.iloc[:, 0], name="response")
    assert_series_equal(expected_response, inferences["series"]["response"])

    expected_cumsum = pd.Series(np.cumsum(expected_response), name="cum_response")

    assert_series_equal(expected_cumsum, inferences["series"]["cum_response"])

    predictor = trained_model.get_prediction()
    forecaster = trained_model.get_forecast(df_post, alpha=alpha)

    pre_pred = predictor.predicted_mean
    post_pred = forecaster.predicted_mean

    point_pred = np.concatenate([pre_pred, post_pred])

    expected_point_pred = pd.Series(point_pred, name="point_pred")
    assert_series_equal(expected_point_pred, inferences["series"]["point_pred"])

    pre_ci = pd.DataFrame(predictor.conf_int(alpha=alpha))
    pre_ci.index = df_pre.index
    post_ci = pd.DataFrame(forecaster.conf_int(alpha=alpha))
    post_ci.index = df_post.index

    ci = pd.concat([pre_ci, post_ci])

    expected_pred_upper = ci.iloc[:, 1]
    expected_pred_upper = expected_pred_upper.rename("point_pred_upper")
    expected_pred_lower = ci.iloc[:, 0]
    expected_pred_lower = expected_pred_lower.rename("point_pred_lower")

    assert_series_equal(expected_pred_upper, inferences["series"]["point_pred_upper"])
    assert_series_equal(expected_pred_lower, inferences["series"]["point_pred_lower"])

    expected_cum_pred = pd.Series(np.cumsum(point_pred), name="cum_pred")
    assert_series_equal(expected_cum_pred, inferences["series"]["cum_pred"])

    expected_cum_pred_lower = pd.Series(
        np.cumsum(expected_pred_lower), name="cum_pred_lower"
    )
    assert_series_equal(expected_cum_pred_lower, inferences["series"]["cum_pred_lower"])

    expected_cum_pred_upper = pd.Series(
        np.cumsum(expected_pred_upper), name="cum_pred_upper"
    )
    assert_series_equal(expected_cum_pred_upper, inferences["series"]["cum_pred_upper"])

    expected_point_effect = pd.Series(
        expected_response - expected_point_pred, name="point_effect"
    )
    assert_series_equal(expected_point_effect, inferences["series"]["point_effect"])

    expected_point_effect_lower = pd.Series(
        expected_response - expected_pred_lower, name="point_effect_lower"
    )
    assert_series_equal(
        expected_point_effect_lower, inferences["series"]["point_effect_lower"]
    )

    expected_point_effect_upper = pd.Series(
        expected_response - expected_pred_upper, name="point_effect_upper"
    )
    assert_series_equal(
        expected_point_effect_upper, inferences["series"]["point_effect_upper"]
    )

    expected_cum_effect = pd.Series(
        np.concatenate(
            (
                np.zeros(len(df_pre)),
                np.cumsum(expected_point_effect.iloc[len(df_pre) :]),
            )
        ),
        name="cum_effect",
    )
    assert_series_equal(expected_cum_effect, inferences["series"]["cum_effect"])

    expected_cum_effect_lower = pd.Series(
        np.concatenate(
            (
                np.zeros(len(df_pre)),
                np.cumsum(expected_point_effect_lower.iloc[len(df_pre) :]),
            )
        ),
        name="cum_effect_lower",
    )
    assert_series_equal(
        expected_cum_effect_lower, inferences["series"]["cum_effect_lower"]
    )

    expected_cum_effect_upper = pd.Series(
        np.concatenate(
            (
                np.zeros(len(df_pre)),
                np.cumsum(expected_point_effect_upper.iloc[len(df_pre) :]),
            )
        ),
        name="cum_effect_upper",
    )
    assert_series_equal(
        expected_cum_effect_upper, inferences["series"]["cum_effect_upper"]
    )


def test_compile_inferences_w_post_period_response(
    data, pre_period, post_period, trained_model
):

    df_pre = data.loc[pre_period[0] : pre_period[1], :]
    df_post = data.loc[post_period[0] : post_period[1], :]

    post_period_response = df_post.loc[post_period[0] : post_period[1]]

    X = df_post.iloc[:, 1:]
    y = X.copy()
    y[:] = np.nan

    df_post = pd.DataFrame(np.concatenate([y, X], axis=1))
    data_index = data.index
    data = pd.concat([df_pre, df_post], axis=0)
    data.index = data_index

    alpha = 0.05
    orig_std_params = (0.0, 1.0)
    estimation = "MLE"

    inferences = compile_inferences(
        trained_model,
        data,
        df_pre,
        None,
        post_period_response,
        alpha,
        orig_std_params,
        estimation,
    )

    expected_response = pd.Series(data.iloc[:, 0], name="response")
    assert_series_equal(expected_response, inferences["series"]["response"])

    expected_cumsum = pd.Series(np.cumsum(expected_response), name="cum_response")

    assert_series_equal(expected_cumsum, inferences["series"]["cum_response"])

    predictor = trained_model.get_prediction(end=len(df_pre) - 1)
    forecaster = trained_model.get_prediction(start=len(df_pre))

    pre_pred = predictor.predicted_mean
    post_pred = forecaster.predicted_mean

    point_pred = np.concatenate([pre_pred, post_pred])

    expected_point_pred = pd.Series(point_pred, name="point_pred")
    assert_series_equal(expected_point_pred, inferences["series"]["point_pred"])

    pre_ci = pd.DataFrame(predictor.conf_int(alpha=alpha))
    pre_ci.index = df_pre.index
    post_ci = pd.DataFrame(forecaster.conf_int(alpha=alpha))
    post_ci.index = df_post.index

    ci = pd.concat([pre_ci, post_ci])

    expected_pred_upper = ci.iloc[:, 1]
    expected_pred_upper = expected_pred_upper.rename("point_pred_upper")
    expected_pred_upper.index = data.index

    expected_pred_lower = ci.iloc[:, 0]
    expected_pred_lower = expected_pred_lower.rename("point_pred_lower")
    expected_pred_lower.index = data.index

    assert_series_equal(expected_pred_upper, inferences["series"]["point_pred_upper"])
    assert_series_equal(expected_pred_lower, inferences["series"]["point_pred_lower"])

    expected_cum_pred = pd.Series(np.cumsum(point_pred), name="cum_pred")
    assert_series_equal(expected_cum_pred, inferences["series"]["cum_pred"])

    expected_cum_pred_lower = pd.Series(
        np.cumsum(expected_pred_lower), name="cum_pred_lower"
    )
    assert_series_equal(expected_cum_pred_lower, inferences["series"]["cum_pred_lower"])

    expected_cum_pred_upper = pd.Series(
        np.cumsum(expected_pred_upper), name="cum_pred_upper"
    )
    assert_series_equal(expected_cum_pred_upper, inferences["series"]["cum_pred_upper"])

    expected_point_effect = pd.Series(
        expected_response - expected_point_pred, name="point_effect"
    )
    assert_series_equal(expected_point_effect, inferences["series"]["point_effect"])

    expected_point_effect_lower = pd.Series(
        expected_response - expected_pred_lower, name="point_effect_lower"
    )
    assert_series_equal(
        expected_point_effect_lower, inferences["series"]["point_effect_lower"]
    )

    expected_point_effect_upper = pd.Series(
        expected_response - expected_pred_upper, name="point_effect_upper"
    )
    assert_series_equal(
        expected_point_effect_upper, inferences["series"]["point_effect_upper"]
    )

    expected_cum_effect = pd.Series(
        np.concatenate(
            (
                np.zeros(len(df_pre)),
                np.cumsum(expected_point_effect.iloc[len(df_pre) :]),
            )
        ),
        name="cum_effect",
    )
    assert_series_equal(expected_cum_effect, inferences["series"]["cum_effect"])

    expected_cum_effect_lower = pd.Series(
        np.concatenate(
            (
                np.zeros(len(df_pre)),
                np.cumsum(expected_point_effect_lower.iloc[len(df_pre) :]),
            )
        ),
        name="cum_effect_lower",
    )
    assert_series_equal(
        expected_cum_effect_lower, inferences["series"]["cum_effect_lower"]
    )

    expected_cum_effect_upper = pd.Series(
        np.concatenate(
            (
                np.zeros(len(df_pre)),
                np.cumsum(expected_point_effect_upper.iloc[len(df_pre) :]),
            )
        ),
        name="cum_effect_upper",
    )
    assert_series_equal(
        expected_cum_effect_upper, inferences["series"]["cum_effect_upper"]
    )



"""Tests for misc module."""


import unittest.mock as mock
import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal
from numpy.testing import assert_almost_equal
import pytest

import causalimpact


standardize = causalimpact.misc.standardize_all_variables
unstandardize = causalimpact.misc.unstandardize
df_print = causalimpact.misc.df_print


def test_basic_standardize():
    pre_period = [0, 2]
    post_period = [3, 4]

    data = {"c1": [1, 4, 8, 9, 10], "c2": [4, 8, 12, 16, 20]}
    data = pd.DataFrame(data)

    result = standardize(data, pre_period, post_period)
    assert_almost_equal(np.zeros((2)), np.mean(result["data_pre"].values, axis=0))

    assert_almost_equal(np.ones((2)), np.std(result["data_pre"].values, axis=0))
    assert len(result["data_pre"]) == pre_period[-1] + 1


def test_standardize_returns_expected_types():
    pre_period = [0, 4]
    post_period = [5, 5]

    data = [-1, 0.1, 1, 2, np.nan, 3]
    data = pd.DataFrame(data)

    result = standardize(data, pre_period, post_period)

    assert isinstance(result, dict)
    assert set(result.keys()) == {"data_pre", "data_post", "orig_std_params"}

    assert len(result["data_pre"]) == pre_period[-1] + 1
    assert_frame_equal(
        unstandardize(result["data_pre"], result["orig_std_params"]),
        pd.DataFrame(data[:5]),
    )


def test_standardize_w_distinct_inputs():
    test_data = [[1], [1, 1, 1], [1, np.nan, 3], pd.DataFrame([10, 20, 30])]

    test_data = [pd.DataFrame(data, dtype="float") for data in test_data]

    for data in test_data:
        result = standardize(
            data,
            pre_period=[0, len(data) + 1],
            post_period=[len(data) + 1, len(data) + 1],
        )

        pd.util.testing.assert_frame_equal(
            unstandardize(result["data_pre"], result["orig_std_params"]), data
        )


def test_standardize_raises_w_bad_input():
    with pytest.raises(ValueError):
        standardize("text", 1, 2)

    with pytest.raises(ValueError):
        standardize(pd.DataFrame([1, 2]), 1, 2)


def test_unstandardize():
    data = np.array([-1.16247639, -0.11624764, 1.27872403])
    orig_std_params = (4.3333333, 2.8674417556)
    original_data = unstandardize(data, orig_std_params)

    assert_almost_equal(original_data.values, np.array([[1.0, 4.0, 8.0]]).T)


def test_df_print():
    data_mock = mock.Mock()
    df_print(data_mock)
    data_mock.assert_not_called()

    df_print(data_mock, path="path")
    data_mock.to_csv.assert_called_once_with("path")



"""Unit Tests for model module"""


import pytest
import unittest.mock as mock
import numpy as np
from numpy.testing import assert_array_equal
import pandas as pd

import causalimpact
from causalimpact.model import ModelResults
from statsmodels.tsa.statespace.structural import UnobservedComponents

observations_validate = causalimpact.model.observations_ill_conditioned
construct_model = causalimpact.model.construct_model
model_fit = causalimpact.model.model_fit


def test_raises_when_y_is_None():
    with pytest.raises(ValueError) as excinfo:
        observations_validate(None)
    assert str(excinfo.value) == "y cannot be None"


def test_raises_when_y_has_len_1():
    with pytest.raises(ValueError) as excinfo:
        observations_validate([1])
    assert str(excinfo.value) == "y must have len > 1"


def test_raises_when_y_is_all_nan():
    with pytest.raises(ValueError) as excinfo:
        observations_validate([np.nan, np.nan])
    assert str(excinfo.value) == (
        "Aborting inference due to input series " "being all null."
    )


def test_raises_when_y_have_just_2_values():
    with pytest.raises(ValueError) as excinfo:
        observations_validate(pd.DataFrame([1, 2]))
    assert str(excinfo.value) == (
        "Aborting inference due to fewer than 3 " "nonnull values in input."
    )


def test_raises_when_y_is_constant():
    with pytest.raises(ValueError) as excinfo:
        observations_validate(pd.Series([1, 1, 1, 1, 1]))
    assert str(excinfo.value) == (
        "Aborting inference due to input series " "being constant"
    )


def test_model_constructor():
    data = pd.DataFrame(np.random.randn(200, 2))
    model = construct_model(data)
    assert_array_equal(model.data.endog, data.iloc[:, 0].values)
    assert model.irregular
    assert model.k_exog == data.shape[1] - 1
    assert model.level
    assert_array_equal(
        model.exog, data.iloc[:, 1].values.reshape(-1, data.shape[1] - 1)
    )


def test_model_constructor_w_just_endog():
    data = pd.DataFrame(np.random.randn(200, 1))
    model = construct_model(data)
    assert_array_equal(model.data.endog, data.iloc[:, 0].values)
    assert model.irregular
    assert model.k_exog == data.shape[1] - 1
    assert model.level
    assert not model.exog


def test_model_fit_with_mle():
    model_mock = mock.Mock()

    model_results = model_fit(model_mock, "MLE", {"niter": 50})
    model_mock.fit.assert_called_once_with(maxiter=50)
    assert isinstance(model_results, ModelResults)


def test_model_fit_with_pymc():
    model_mock = UnobservedComponents([1, 2, 3, 4], level="llevel")
    model_results = model_fit(model_mock, "pymc", {"ndraws": 2, "nburn": 1})
    assert isinstance(model_results, ModelResults)



import numpy as np
import pandas as pd
from causalimpact.misc import unstandardize


def compile_inferences(
    results,
    data,
    df_pre,
    df_post,
    post_period_response,
    alpha,
    orig_std_params,
    estimation,
):
    """Compiles inferences to make predictions for post intervention
    period.

    Args:
        results: trained UnobservedComponents model from statsmodels package.
        data: pd.DataFrame pre and post-intervention data containing y and X.
        df_pre: pd.DataFrame pre intervention data
        df_post: pd.DataFrame post intervention data
        post_period_response: pd.DataFrame used when the model trained is not
            default one but a customized instead. In this case,
            ``df_post`` is None.
        alpha: float significance level for confidence intervals.
        orig_std_params: tuple of floats where first value is the mean and
            second value is standard deviation used for standardizing data.
        estimation: str to choose fitting method. "MLE" as default

    Returns:
        dict containing all data related to the inference process.
    """
    # Compute point predictions of counterfactual (in standardized space)
    if df_post is not None:
        # returns pre-period predictions (in-sample)
        predict = results.get_prediction()
        # returns post-period forecast
        forecast = results.get_forecast(df_post, alpha=alpha)
    else:
        pre_len = results.model_nobs - len(post_period_response)

        predict = results.get_prediction(end=pre_len - 1)
        forecast = results.get_prediction(start=pre_len)

        df_post = post_period_response
        df_post.index = pd.core.indexes.range.RangeIndex(
            start=pre_len, stop=pre_len + len(df_post), step=1
        )

    # Compile summary statistics (in original space)
    pre_pred = unstandardize(predict.predicted_mean, orig_std_params)
    pre_pred.index = df_pre.index

    post_pred = unstandardize(forecast.predicted_mean, orig_std_params)
    post_pred.index = df_post.index

    point_pred = pd.concat([pre_pred, post_pred])

    pre_ci = unstandardize(predict.conf_int(alpha=alpha), orig_std_params)
    pre_ci.index = df_pre.index

    post_ci = unstandardize(forecast.conf_int(alpha=alpha), orig_std_params)

    post_ci.index = df_post.index
    ci = pd.concat([pre_ci, post_ci])
    point_pred_lower = ci.iloc[:, 0].to_frame()
    point_pred_upper = ci.iloc[:, 1].to_frame()

    response = data.iloc[:, 0]
    response_index = data.index

    response = pd.DataFrame(response)

    cum_response = np.cumsum(response)
    cum_pred = np.cumsum(point_pred)
    cum_pred_lower = np.cumsum(point_pred_lower)
    cum_pred_upper = np.cumsum(point_pred_upper)

    data = pd.concat(
        [
            point_pred,
            point_pred_lower,
            point_pred_upper,
            cum_pred,
            cum_pred_lower,
            cum_pred_upper,
        ],
        axis=1,
    )

    data = pd.concat([response, cum_response], axis=1).join(data, lsuffix="l")

    data.columns = [
        "response",
        "cum_response",
        "point_pred",
        "point_pred_lower",
        "point_pred_upper",
        "cum_pred",
        "cum_pred_lower",
        "cum_pred_upper",
    ]

    point_effect = (data.response - data.point_pred).to_frame()
    point_effect_lower = (data.response - data.point_pred_lower).to_frame()
    point_effect_upper = (data.response - data.point_pred_upper).to_frame()

    cum_effect = point_effect.copy()
    cum_effect.loc[df_pre.index[0] : df_pre.index[-1]] = 0
    cum_effect = np.cumsum(cum_effect)

    cum_effect_lower = point_effect_lower.copy()
    cum_effect_lower.loc[df_pre.index[0] : df_pre.index[-1]] = 0
    cum_effect_lower = np.cumsum(cum_effect_lower)

    cum_effect_upper = point_effect_upper.copy()
    cum_effect_upper.loc[df_pre.index[0] : df_pre.index[-1]] = 0
    cum_effect_upper = np.cumsum(cum_effect_upper)

    data = pd.concat(
        [
            data,
            point_effect,
            point_effect_lower,
            point_effect_upper,
            cum_effect,
            cum_effect_lower,
            cum_effect_upper,
        ],
        axis=1,
    )

    # Create DataFrame of results
    data.columns = [
        "response",
        "cum_response",
        "point_pred",
        "point_pred_lower",
        "point_pred_upper",
        "cum_pred",
        "cum_pred_lower",
        "cum_pred_upper",
        "point_effect",
        "point_effect_lower",
        "point_effect_upper",
        "cum_effect",
        "cum_effect_lower",
        "cum_effect_upper",
    ]

    data.index = response_index

    series = data

    inferences = {"series": series}
    return inferences



import textwrap
import numpy as np
import pandas as pd
from pandas.api.types import is_list_like

from causalimpact.misc import standardize_all_variables, df_print, get_matplotlib
from causalimpact.model import construct_model, model_fit
from causalimpact.inferences import compile_inferences
import scipy.stats as st


class CausalImpact:
    """CausalImpact() performs causal inference through counterfactual
    predictions using a Bayesian structural time-series model.

    Parameters:
    ----------
        data : pandas dataframe
            the response variable must be in the first column, and any covariates
            in subsequent columns.
        pre_period : list
            A list specifying the first and the last time point of the
            pre-intervention period in the response column. This period can be
            thought of as a training period, used to determine the relationship
            between the response variable and the covariates.
        post_period : list
            A vector specifying the first and the last day of the post-intervention
            period we wish to study. This is the period after the intervention has
            begun whose effect we are interested in. The relationship between
            response variable and covariates, as determined during the pre-period,
            will be used to predict how the response variable should have evolved
            during the post-period had no intervention taken place.
        model_args : dict
            Optional arguments that can be used to adjust the default construction
            of the state-space model used for inference.
            For full control over the model, you can construct your own model using
            the statsmodels package and feed the model into CausalImpact().
        ucm_model : statsmodels.tsa.statespace.structural.UnobservedComponents
            Instead of passing in data and having CausalImpact construct a
            model, it is possible to construct a model yourself using the
            statsmodel package. In this case, omit data, pre_period, and
            post_period. Instead only pass in ucm_model, y_post, alpha (optional).
            The model must have been fitted on data where the response variable was
            set to np.nan during the post-treatment period. The actual observed data
            during this period must then be passed to the function in y_post.
        post_period_response : list | pd.Series | np.Array
            Actual observed data during the post-intervention period. This is required
            if and only if a fitted ucm_model is passed instead of data.
        alpha : float
            Desired tail-area probability for posterior intervals. Defaults to 0.05,
            which will produce central 95% intervals.


        Returns
        -------
        CausalImpact Object

    """

    def __init__(
        self,
        data=None,
        pre_period=None,
        post_period=None,
        model_args=None,
        ucm_model=None,
        post_period_response=None,
        alpha=0.05,
        estimation="MLE",
    ):
        self.series = None
        self.model = {}
        if isinstance(data, pd.DataFrame):
            self.data = data.copy()
        else:
            self.data = data
        self.params = {
            "data": data,
            "pre_period": pre_period,
            "post_period": post_period,
            "model_args": model_args,
            "ucm_model": ucm_model,
            "post_period_response": post_period_response,
            "alpha": alpha,
            "estimation": estimation,
        }
        self.inferences = None
        self.results = None

    def run(self):
        kwargs = self._format_input(
            self.params["data"],
            self.params["pre_period"],
            self.params["post_period"],
            self.params["model_args"],
            self.params["ucm_model"],
            self.params["post_period_response"],
            self.params["alpha"],
        )

        # Depending on input, dispatch to the appropriate Run* method()
        if self.data is not None:
            self._run_with_data(
                kwargs["data"],
                kwargs["pre_period"],
                kwargs["post_period"],
                kwargs["model_args"],
                kwargs["alpha"],
                self.params["estimation"],
            )
        else:
            self._run_with_ucm(
                kwargs["ucm_model"],
                kwargs["post_period_response"],
                kwargs["alpha"],
                kwargs["model_args"],
                self.params["estimation"],
            )

    @staticmethod
    def _format_input_data(data):
        """Check and format the data argument provided to CausalImpact().

        Args:
            data: Pandas DataFrame

        Returns:
            correctly formatted Pandas DataFrame
        """
        # If <data> is a Pandas DataFrame and the first column is 'date',
        # try to convert

        if (
            isinstance(data, pd.DataFrame)
            and isinstance(data.columns[0], str)
            and data.columns[0].lower() in ["date", "time"]
        ):
            data = data.set_index(data.columns[0])

        # Try to convert to Pandas DataFrame
        try:
            data = pd.DataFrame(data)
        except ValueError:
            raise ValueError("could not convert input data to Pandas " + "DataFrame")

        # Must have at least 3 time points
        if len(data.index) < 3:
            raise ValueError("data must have at least 3 time points")

        # Must not have NA in covariates (if any)
        if len(data.columns) >= 2 and pd.isnull(data.iloc[:, 1:]).any(axis=None):
            raise ValueError("covariates must not contain null values")

        return data

    @staticmethod
    def _check_periods_are_valid(pre_period, post_period):
        if not isinstance(pre_period, list) or not isinstance(post_period, list):
            raise ValueError("pre_period and post_period must both be lists")
        if len(pre_period) != 2 or len(post_period) != 2:
            raise ValueError("pre_period and post_period must both be of " + "length 2")
        if pd.isnull(pre_period).any(axis=None) or pd.isnull(post_period).any(
            axis=None
        ):
            raise ValueError(
                "pre_period and post period must not contain " + "null values"
            )

    @staticmethod
    def _align_periods_dtypes(pre_period, post_period, data):
        """align the dtypes of the pre_period and post_period to the data index.

        Args:
            pre_period: two-element list
            post_period: two-element list
            data: already-checked Pandas DataFrame, for reference only
        """
        pre_dtype = np.array(pre_period).dtype
        post_dtype = np.array(post_period).dtype
        # if index is datetime then convert pre and post to datetimes
        if isinstance(data.index, pd.core.indexes.datetimes.DatetimeIndex):
            pre_period = [pd.to_datetime(date) for date in pre_period]
            post_period = [pd.to_datetime(date) for date in post_period]
            pd.core.dtypes.common.is_datetime_or_timedelta_dtype(pre_period)
        # if index is not datetime then error if datetime pre and post is passed
        elif pd.core.dtypes.common.is_datetime_or_timedelta_dtype(
            pd.Series(pre_period)
        ) or pd.core.dtypes.common.is_datetime_or_timedelta_dtype(
            pd.Series(post_period)
        ):
            raise ValueError(
                "pre_period ("
                + pre_dtype.name
                + ") and post_period ("
                + post_dtype.name
                + ") should have the same class as the "
                + "time points in the data ("
                + data.index.dtype.name
                + ")"
            )
        # if index is int
        elif pd.api.types.is_int64_dtype(data.index):
            pre_period = [int(elem) for elem in pre_period]
            post_period = [int(elem) for elem in post_period]
        # if index is int
        elif pd.api.types.is_float_dtype(data.index):
            pre_period = [float(elem) for elem in pre_period]
            post_period = [float(elem) for elem in post_period]
        # if index is string
        elif pd.api.types.is_string_dtype(data.index):
            if pd.api.types.is_numeric_dtype(
                np.array(pre_period)
            ) or pd.api.types.is_numeric_dtype(np.array(post_period)):
                raise ValueError(
                    "pre_period ("
                    + pre_dtype.name
                    + ") and post_period ("
                    + post_dtype.name
                    + ") should have the same class as the "
                    + "time points in the data ("
                    + data.index.dtype.name
                    + ")"
                )
            else:
                pre_period = [str(idx) for idx in pre_period]
                post_period = [str(idx) for idx in post_period]
        else:
            raise ValueError(
                "pre_period ("
                + pre_dtype.name
                + ") and post_period ("
                + post_dtype.name
                + ") should have the same class as the "
                + "time points in the data ("
                + data.index.dtype.name
                + ")"
            )
        return [pre_period, post_period]

    def _format_input_prepost(self, pre_period, post_period, data):
        """Check and format the pre_period and post_period input arguments.

        Args:
            pre_period: two-element list
            post_period: two-element list
            data: already-checked Pandas DataFrame, for reference only
        """
        self._check_periods_are_valid(pre_period, post_period)

        pre_period, post_period = self._align_periods_dtypes(
            pre_period, post_period, data
        )

        if pre_period[1] > post_period[0]:
            raise ValueError(
                "post period must start at least 1 observation"
                + " after the end of the pre_period"
            )

        if isinstance(data.index, pd.RangeIndex):
            loc3 = post_period[0]
            loc4 = post_period[1]
        else:
            loc3 = data.index.get_loc(post_period[0])
            loc4 = data.index.get_loc(post_period[1])

        if loc4 < loc3:
            raise ValueError(
                "post_period[1] must not be earlier than " + "post_period[0]"
            )

        if pre_period[0] < data.index.min():
            pre_period[0] = data.index.min()
        if post_period[1] > data.index.max():
            post_period[1] = data.index.max()
        return {"pre_period": pre_period, "post_period": post_period}

    @staticmethod
    def _check_valid_args_combo(args):
        data_model_args = [True, True, True, False, False]
        ucm_model_args = [False, False, False, True, True]

        if np.any(pd.isnull(args) != data_model_args) and np.any(
            pd.isnull(args) != ucm_model_args
        ):
            raise SyntaxError(
                "Must either provide ``data``, ``pre_period``"
                + " ,``post_period``, ``model_args``"
                " or ``ucm_model" + "and ``post_period_response``"
            )

    @staticmethod
    def _check_valid_alpha(alpha):
        if alpha is None:
            raise ValueError("alpha must not be None")
        if not np.isreal(alpha):
            raise ValueError("alpha must be a real number")
        if np.isnan(alpha):
            raise ValueError("alpha must not be NA")
        if alpha <= 0 or alpha >= 1:
            raise ValueError("alpha must be between 0 and 1")

    def _format_input(
        self,
        data,
        pre_period,
        post_period,
        model_args,
        ucm_model,
        post_period_response,
        alpha,
    ):
        """Check and format all input arguments supplied to CausalImpact().
           See the documentation of CausalImpact() for details

        Args:
            data:                 Pandas DataFrame or data frame
            pre_period:           beginning and end of pre-period
            post_period:          beginning and end of post-period
            model_args:           dict of additional arguments for the model
            ucm_model:            UnobservedComponents model (instead of data)
            post_period_response: observed response in the post-period
            alpha:                tail-area for posterior intervals
            estimation:           method of estimation for model fitting

        Returns:
            list of checked (and possibly reformatted) input arguments
        """
        from statsmodels.tsa.statespace.structural import UnobservedComponents

        # Check that a consistent set of variables has been provided
        args = [data, pre_period, post_period, ucm_model, post_period_response]

        self._check_valid_args_combo(args)

        # Check <data> and convert to Pandas DataFrame, with rows
        # representing time points
        if data is not None:
            data = self._format_input_data(data)

        # Check <pre_period> and <post_period>
        if data is not None:
            checked = self._format_input_prepost(pre_period, post_period, data)
            pre_period = checked["pre_period"]
            post_period = checked["post_period"]
            self.params["pre_period"] = pre_period
            self.params["post_period"] = post_period

        # Parse <model_args>, fill gaps using <_defaults>

        _defaults = {
            "ndraws": 1000,
            "nburn": 100,
            "niter": 1000,
            "standardize_data": True,
            "prior_level_sd": 0.01,
            "nseasons": 1,
            "season_duration": 1,
            "dynamic_regression": False,
        }

        if model_args is None:
            model_args = _defaults
        else:
            missing = [key for key in _defaults if key not in model_args]
            for arg in missing:
                model_args[arg] = _defaults[arg]

        # Check <standardize_data>
        if not isinstance(model_args["standardize_data"], bool):
            raise ValueError("model_args.standardize_data must be a" + " boolean value")

        # Check <ucm_model>
        if ucm_model is not None and not isinstance(ucm_model, UnobservedComponents):
            raise ValueError(
                "ucm_model must be an object of class "
                "statsmodels.tsa.statespace.structural.UnobservedComponents "
                "instead received " + str(type(ucm_model))[8:-2]
            )

        # Check <post_period_response>
        if ucm_model is not None:
            if not is_list_like(post_period_response):
                raise ValueError("post_period_response must be list-like")
            if np.array(post_period_response).dtype.num == 17:
                raise ValueError(
                    "post_period_response should not be" + " datetime values"
                )
            if not np.all(np.isreal(post_period_response)):
                raise ValueError(
                    "post_period_response must contain all" + " real values"
                )

        # Check <alpha>
        self._check_valid_alpha(alpha)

        # Return updated arguments
        kwargs = {
            "data": data,
            "pre_period": pre_period,
            "post_period": post_period,
            "model_args": model_args,
            "ucm_model": ucm_model,
            "post_period_response": post_period_response,
            "alpha": alpha,
        }
        return kwargs

    def _run_with_data(
        self, data, pre_period, post_period, model_args, alpha, estimation
    ):
        # Zoom in on data in modeling range
        if data.shape[1] == 1:  # no exogenous values provided
            raise ValueError("data contains no exogenous variables")
        data_modeling = data.copy()

        df_pre = data_modeling.loc[pre_period[0] : pre_period[1], :]
        df_post = data_modeling.loc[post_period[0] : post_period[1], :]

        # Standardize all variables
        orig_std_params = (0, 1)
        if model_args["standardize_data"]:
            sd_results = standardize_all_variables(
                data_modeling, pre_period, post_period
            )
            df_pre = sd_results["data_pre"]
            df_post = sd_results["data_post"]
            orig_std_params = sd_results["orig_std_params"]

        # Construct model and perform inference
        model = construct_model(df_pre, model_args)
        self.model = model

        model_results = model_fit(model, estimation, model_args)

        inferences = compile_inferences(
            model_results,
            data,
            df_pre,
            df_post,
            None,
            alpha,
            orig_std_params,
            estimation,
        )

        # "append" to 'CausalImpact' object
        self.inferences = inferences["series"]
        self.results = model_results

    def _run_with_ucm(
        self, ucm_model, post_period_response, alpha, model_args, estimation
    ):
        """Runs an impact analysis on top of a ucm model.

        Args:
          ucm_model: Model as returned by UnobservedComponents(),
                     in which the data during the post-period was set to NA
          post_period_response: observed data during the post-intervention
                                period
          alpha: tail-probabilities of posterior intervals"""

        df_pre = ucm_model.data.orig_endog[: -len(post_period_response)]
        df_pre = pd.DataFrame(df_pre)

        post_period_response = pd.DataFrame(post_period_response)

        data = pd.DataFrame(
            np.concatenate([df_pre.values, post_period_response.values])
        )

        orig_std_params = (0, 1)

        model_results = model_fit(ucm_model, estimation, model_args)

        # Compile posterior inferences
        inferences = compile_inferences(
            model_results,
            data,
            df_pre,
            None,
            post_period_response,
            alpha,
            orig_std_params,
            estimation,
        )

        obs_inter = model_results.model_nobs - len(post_period_response)

        self.params["pre_period"] = [0, obs_inter - 1]
        self.params["post_period"] = [obs_inter, -1]
        self.data = pd.concat([df_pre, post_period_response])
        self.inferences = inferences["series"]
        self.results = model_results

    @staticmethod
    def _print_report(
        mean_pred_fmt,
        mean_resp_fmt,
        mean_lower_fmt,
        mean_upper_fmt,
        abs_effect_fmt,
        abs_effect_upper_fmt,
        abs_effect_lower_fmt,
        rel_effect_fmt,
        rel_effect_upper_fmt,
        rel_effect_lower_fmt,
        cum_resp_fmt,
        cum_pred_fmt,
        cum_lower_fmt,
        cum_upper_fmt,
        confidence,
        cum_rel_effect_lower,
        cum_rel_effect_upper,
        cum_rel_effect,
        width,
        p_value,
        alpha,
    ):
        sig = not (cum_rel_effect_lower < 0 < cum_rel_effect_upper)
        pos = cum_rel_effect > 0
        # Summarize averages
        stmt = textwrap.dedent(
            """During the post-intervention period, the response
            variable had an average value of
            approx. {mean_resp}.
                        """.format(
                mean_resp=mean_resp_fmt
            )
        )
        if sig:
            stmt += " By contrast, in "
        else:
            stmt += " In "

        stmt += textwrap.dedent(
            """
                the absence of an intervention, we would have
                expected an average response of {mean_pred}. The
                {confidence} interval of this counterfactual
                prediction is [{mean_lower}, {mean_upper}].
                Subtracting this prediction from the observed
                response yields an estimate of the causal effect
                the intervention had on the response variable.
                This effect is {abs_effect} with a
                {confidence} interval of [{abs_lower},
                {abs_upper}]. For a discussion of the
                significance of this effect,
                see below.
                """.format(
                mean_pred=mean_pred_fmt,
                confidence=confidence,
                mean_lower=mean_lower_fmt,
                mean_upper=mean_upper_fmt,
                abs_effect=abs_effect_fmt,
                abs_upper=abs_effect_upper_fmt,
                abs_lower=abs_effect_lower_fmt,
            )
        )
        # Summarize sums
        stmt2 = textwrap.dedent(
            """
                Summing up the individual data points during the
                post-intervention period (which can only sometimes be
                meaningfully interpreted), the response variable had an
                overall value of {cum_resp}.
                """.format(
                cum_resp=cum_resp_fmt
            )
        )
        if sig:
            stmt2 += " By contrast, had "
        else:
            stmt2 += " Had "

        stmt2 += textwrap.dedent(
            """
                the intervention not taken place, we would have expected
                a sum of {cum_pred}. The {confidence} interval of this
                prediction is [{cum_pred_lower}, {cum_pred_upper}]
                """.format(
                cum_pred=cum_pred_fmt,
                confidence=confidence,
                cum_pred_lower=cum_lower_fmt,
                cum_pred_upper=cum_upper_fmt,
            )
        )

        # Summarize relative numbers (in which case row [1] = row [2])
        stmt3 = textwrap.dedent(
            """
                                The above results are given in terms
                                of absolute numbers. In relative terms, the
                                response variable showed
                                """
        )
        if pos:
            stmt3 += " an increase of "
        else:
            stmt3 += " a decrease of "

        stmt3 += textwrap.dedent(
            """
                        {rel_effect}. The {confidence} interval of this
                        percentage is [{rel_effect_lower},
                        {rel_effect_upper}]
                        """.format(
                confidence=confidence,
                rel_effect=rel_effect_fmt,
                rel_effect_lower=rel_effect_lower_fmt,
                rel_effect_upper=rel_effect_upper_fmt,
            )
        )

        # Comment on significance
        if sig and pos:
            stmt4 = textwrap.dedent(
                """
                        This means that the positive effect observed
                        during the intervention period is statistically
                        significant and unlikely to be due to random
                        fluctuations. It should be noted, however, that
                        the question of whether this increase also bears
                        substantive significance can only be answered by
                        comparing the absolute effect {abs_effect} to
                        the original goal of the underlying
                        intervention.
                        """.format(
                    abs_effect=abs_effect_fmt
                )
            )
        elif sig and not pos:
            stmt4 = textwrap.dedent(
                """
                        This  means that the negative effect observed
                        during the intervention period is statistically
                        significant. If the experimenter had expected a
                        positive effect, it is recommended to double-check
                        whether anomalies in the control variables may have
                        caused an overly optimistic expectation of what
                        should have happened in the response variable in the
                        absence of the intervention.
                        """
            )
        elif not sig and pos:
            stmt4 = textwrap.dedent(
                """
                        This means that, although the intervention
                        appears to have caused a positive effect, this
                        effect is not statistically significant when
                        considering the post-intervention period as a whole.
                        Individual days or shorter stretches within the
                        intervention period may of course still have had a
                        significant effect, as indicated whenever the lower
                        limit of the impact time series (lower plot) was
                        above zero.
                        """
            )
        elif not sig and not pos:
            stmt4 = textwrap.dedent(
                """
                        This means that, although it may look as though
                        the intervention has exerted a negative effect on
                        the response variable when considering the
                        intervention period as a whole, this effect is not
                        statistically significant, and so cannot be
                        meaningfully interpreted.
                        """
            )
        if not sig:
            stmt4 += textwrap.dedent(
                """
                        The apparent effect could be the result of random
                        fluctuations that are unrelated to the intervention.
                        This is often the case when the intervention period
                        is very long and includes much of the time when the
                        effect has already worn off. It can also be the case
                        when the intervention period is too short to
                        distinguish the signal from the noise. Finally,
                        failing to find a significant effect can happen when
                        there are not enough control variables or when these
                        variables do not correlate well with the response
                        variable during the learning period."""
            )
        if p_value < alpha:
            stmt5 = textwrap.dedent(
                """The probability of obtaining this effect by
                chance is very small (Bayesian one-sided tail-area
                probability {p}). This means the
                causal effect can be considered statistically
                significant.""".format(
                    p=np.round(p_value, 3)
                )
            )
        else:
            stmt5 = """The probability of obtaining this effect by
                        chance is p = ", round(p, 3), "). This means the effect may
                        be spurious and would generally not be considered
                        statistically significant.""".format()

        print(textwrap.fill(stmt, width=width))
        print("\n")
        print(textwrap.fill(stmt2, width=width))
        print("\n")
        print(textwrap.fill(stmt3, width=width))
        print("\n")
        print(textwrap.fill(stmt4, width=width))
        print("\n")
        print(textwrap.fill(stmt5, width=width))

    def summary(self, output="summary", width=120, path=None):
        """reports a summary of the results

        Parameters
        ----------
        output: str
            can be summary or report. summary outputs a table.
            report outputs a natural language description of the
            findings
        width : int
            line width of the output. Only relevant if output == report
        path : str
            path to output summary to csv. Only relevant if output == summary

        """
        alpha = self.params["alpha"]
        confidence = "{}%".format(int((1 - alpha) * 100))
        post_period = self.params["post_period"]
        post_inf = self.inferences.loc[post_period[0] : post_period[1], :]
        post_point_resp = post_inf.loc[:, "response"]
        post_point_pred = post_inf.loc[:, "point_pred"]
        post_point_upper = post_inf.loc[:, "point_pred_upper"]
        post_point_lower = post_inf.loc[:, "point_pred_lower"]

        mean_resp = post_point_resp.mean()
        mean_resp_fmt = int(mean_resp)
        cum_resp = post_point_resp.sum()
        cum_resp_fmt = int(cum_resp)
        mean_pred = post_point_pred.mean()
        mean_pred_fmt = int(post_point_pred.mean())
        cum_pred = post_point_pred.sum()
        cum_pred_fmt = int(cum_pred)
        mean_lower = post_point_lower.mean()
        mean_lower_fmt = int(mean_lower)
        mean_upper = post_point_upper.mean()
        mean_upper_fmt = int(mean_upper)
        mean_ci_fmt = [mean_lower_fmt, mean_upper_fmt]
        cum_lower = post_point_lower.sum()
        cum_lower_fmt = int(cum_lower)
        cum_upper = post_point_upper.sum()
        cum_upper_fmt = int(cum_upper)
        cum_ci_fmt = [cum_lower_fmt, cum_upper_fmt]

        abs_effect = (post_point_resp - post_point_pred).mean()
        abs_effect_fmt = int(abs_effect)
        cum_abs_effect = (post_point_resp - post_point_pred).sum()
        cum_abs_effect_fmt = int(cum_abs_effect)
        abs_effect_lower = (post_point_resp - post_point_lower).mean()
        abs_effect_lower_fmt = int(abs_effect_lower)
        abs_effect_upper = (post_point_resp - post_point_upper).mean()
        abs_effect_upper_fmt = int(abs_effect_upper)
        abs_effect_ci_fmt = [abs_effect_lower_fmt, abs_effect_upper_fmt]
        cum_abs_lower = (post_point_resp - post_point_lower).sum()
        cum_abs_lower_fmt = int(cum_abs_lower)
        cum_abs_upper = (post_point_resp - post_point_upper).sum()
        cum_abs_upper_fmt = int(cum_abs_upper)
        cum_abs_effect_ci_fmt = [cum_abs_lower_fmt, cum_abs_upper_fmt]

        rel_effect = abs_effect / mean_pred * 100
        rel_effect_fmt = "{:.1f}%".format(rel_effect)
        cum_rel_effect = cum_abs_effect / cum_pred * 100
        cum_rel_effect_fmt = "{:.1f}%".format(cum_rel_effect)
        rel_effect_lower = abs_effect_lower / mean_pred * 100
        rel_effect_lower_fmt = "{:.1f}%".format(rel_effect_lower)
        rel_effect_upper = abs_effect_upper / mean_pred * 100
        rel_effect_upper_fmt = "{:.1f}%".format(rel_effect_upper)
        rel_effect_ci_fmt = [rel_effect_lower_fmt, rel_effect_upper_fmt]
        cum_rel_effect_lower = cum_abs_lower / cum_pred * 100
        cum_rel_effect_lower_fmt = "{:.1f}%".format(cum_rel_effect_lower)
        cum_rel_effect_upper = cum_abs_upper / cum_pred * 100
        cum_rel_effect_upper_fmt = "{:.1f}%".format(cum_rel_effect_upper)
        cum_rel_effect_ci_fmt = [cum_rel_effect_lower_fmt, cum_rel_effect_upper_fmt]

        # assuming approximately normal distribution
        # calculate standard deviation from the 95% conf interval
        std_pred = (
            mean_upper - mean_pred
        ) / 1.96  # from mean_upper = mean_pred + 1.96 * std
        # calculate z score
        z_score = (0 - mean_pred) / std_pred
        # pvalue from zscore
        p_value = st.norm.cdf(z_score)
        prob_causal = 1 - p_value
        p_value_perc = p_value * 100
        prob_causal_perc = prob_causal * 100

        if output == "summary":
            # Posterior inference {CausalImpact}
            summary = [
                [mean_resp_fmt, cum_resp_fmt],
                [mean_pred_fmt, cum_pred_fmt],
                [mean_ci_fmt, cum_ci_fmt],
                [" ", " "],
                [abs_effect_fmt, cum_abs_effect_fmt],
                [abs_effect_ci_fmt, cum_abs_effect_ci_fmt],
                [" ", " "],
                [rel_effect_fmt, cum_rel_effect_fmt],
                [rel_effect_ci_fmt, cum_rel_effect_ci_fmt],
                [" ", " "],
                ["{:.1f}%".format(p_value_perc), " "],
                ["{:.1f}%".format(prob_causal_perc), " "],
            ]
            summary = pd.DataFrame(
                summary,
                columns=["Average", "Cumulative"],
                index=[
                    "Actual",
                    "Predicted",
                    "95% CI",
                    " ",
                    "Absolute Effect",
                    "95% CI",
                    " ",
                    "Relative Effect",
                    "95% CI",
                    " ",
                    "P-value",
                    "Prob. of Causal Effect",
                ],
            )
            df_print(summary, path)
        elif output == "report":
            self._print_report(
                mean_pred_fmt,
                mean_resp_fmt,
                mean_lower_fmt,
                mean_upper_fmt,
                abs_effect_fmt,
                abs_effect_upper_fmt,
                abs_effect_lower_fmt,
                rel_effect_fmt,
                rel_effect_upper_fmt,
                rel_effect_lower_fmt,
                cum_resp_fmt,
                cum_pred_fmt,
                cum_lower_fmt,
                cum_upper_fmt,
                confidence,
                cum_rel_effect_lower,
                cum_rel_effect_upper,
                cum_rel_effect,
                width,
                p_value,
                alpha,
            )
        else:
            raise ValueError(
                "Output argument must be either 'summary' " + "or 'report'"
            )

    def plot(
        self,
        panels=None,
        figsize=(15, 12),
        fname=None,
    ):
        if panels is None:
            panels = ["original", "pointwise", "cumulative"]
        plt = get_matplotlib()
        fig = plt.figure(figsize=figsize)

        data_inter = self.params["pre_period"][1]
        if isinstance(data_inter, pd.DatetimeIndex):
            data_inter = pd.Timestamp(data_inter)

        inferences = self.inferences.iloc[1:, :]

        # Observation and regression components
        if "original" in panels:
            ax1 = plt.subplot(3, 1, 1)
            plt.plot(inferences.point_pred, "r--", linewidth=2, label="model")
            plt.plot(inferences.response, "k", linewidth=2, label="endog")

            plt.axvline(data_inter, c="k", linestyle="--")

            plt.fill_between(
                inferences.index,
                inferences.point_pred_lower,
                inferences.point_pred_upper,
                facecolor="gray",
                interpolate=True,
                alpha=0.25,
            )
            plt.setp(ax1.get_xticklabels(), visible=False)
            plt.legend(loc="upper left")
            plt.title("Observation vs prediction")

        if "pointwise" in panels:
            # Pointwise difference
            if "ax1" in locals():
                ax2 = plt.subplot(312, sharex=ax1)
            else:
                ax2 = plt.subplot(312)
            lift = inferences.point_effect
            plt.plot(lift, "r--", linewidth=2)
            plt.plot(self.data.index, np.zeros(self.data.shape[0]), "g-", linewidth=2)
            plt.axvline(data_inter, c="k", linestyle="--")

            lift_lower = inferences.point_effect_lower
            lift_upper = inferences.point_effect_upper

            plt.fill_between(
                inferences.index,
                lift_lower,
                lift_upper,
                facecolor="gray",
                interpolate=True,
                alpha=0.25,
            )
            plt.setp(ax2.get_xticklabels(), visible=False)
            plt.title("Difference")

        # Cumulative impact
        if "cumulative" in panels:
            if "ax1" in locals():
                plt.subplot(313, sharex=ax1)
            elif "ax2" in locals():
                plt.subplot(313, sharex=ax2)
            else:
                plt.subplot(313)
            plt.plot(
                inferences.index,
                inferences.cum_effect,
                "r--",
                linewidth=2,
            )

            plt.plot(self.data.index, np.zeros(self.data.shape[0]), "g-", linewidth=2)
            plt.axvline(data_inter, c="k", linestyle="--")

            plt.fill_between(
                inferences.index,
                inferences.cum_effect_lower,
                inferences.cum_effect_upper,
                facecolor="gray",
                interpolate=True,
                alpha=0.25,
            )
            plt.axis([inferences.index[0], inferences.index[-1], None, None])

            plt.title("Cumulative Impact")
        plt.xlabel("$T$")
        if fname is None:
            plt.show()
        else:
            fig.savefig(fname, bbox_inches="tight")
            plt.close(fig)



"""Unit Tests for analysis module."""


import os
import re
from tempfile import mkdtemp
import unittest.mock as mock
import pytest
import numpy as np
from numpy.testing import assert_array_equal
import pandas as pd
from pandas.testing import assert_frame_equal
from statsmodels.tsa.statespace.structural import UnobservedComponents as UCM
from causalimpact import CausalImpact
from unittest import TestCase


@pytest.fixture()
def data():
    return pd.DataFrame(np.random.randn(200, 3), columns=["y", "x1", "x2"])


@pytest.fixture()
def summary_report_filename(FIXTURES_FOLDER):
    return os.path.join(FIXTURES_FOLDER, "analysis", "summary_report_output.txt")


@pytest.fixture()
def expected_columns():
    return [
        "response",
        "cum_response",
        "point_pred",
        "point_pred_lower",
        "point_pred_upper",
        "cum_pred",
        "cum_pred_lower",
        "cum_pred_upper",
        "point_effect",
        "point_effect_lower",
        "point_effect_upper",
        "cum_effect",
        "cum_effect_lower",
        "cum_effect_upper",
    ]


@pytest.fixture()
def inference_input():
    return {
        "response": np.array([1.0, 2.0, 3.0, 4.0]),
        "point_pred": np.array([1.1, 2.2, 3.1, 4.1]),
        "point_pred_upper": np.array([1.5, 2.6, 3.4, 4.4]),
        "point_pred_lower": np.array([1.0, 2.0, 3.0, 4.0]),
    }


@pytest.fixture()
def pre_period():
    return [0, 100]


@pytest.fixture()
def post_period():
    return [101, 199]


@pytest.fixture()
def ucm_model(data, post_period):
    data_modeling = data.copy()
    data_modeling[post_period[0] : post_period[1] + 1] = np.nan
    return UCM(endog=data_modeling.iloc[:, 0].values, level="llevel")


@pytest.fixture()
def impact_ucm(ucm_model):
    post_period_response = np.random.randn(100)
    return CausalImpact(ucm_model=ucm_model, post_period_response=post_period_response)


@pytest.fixture()
def causal_impact(data, pre_period, post_period):
    model_args = {"niter": 123}
    return CausalImpact(data, pre_period, post_period, model_args)


class TestFormatInput:
    @staticmethod
    def test_input_default_model(causal_impact):
        expected = {
            "data": causal_impact.params["data"],
            "pre_period": causal_impact.params["pre_period"],
            "post_period": causal_impact.params["post_period"],
            "model_args": causal_impact.params["model_args"],
            "ucm_model": None,
            "post_period_response": None,
            "alpha": causal_impact.params["alpha"],
        }

        result = causal_impact._format_input(
            causal_impact.params["data"],
            causal_impact.params["pre_period"],
            causal_impact.params["post_period"],
            causal_impact.params["model_args"],
            None,
            None,
            causal_impact.params["alpha"],
        )

        result_data = result["data"]
        expected_data = expected["data"]
        assert_frame_equal(result_data, expected_data)

        result_model_args = result["model_args"]
        expected_model_args = expected["model_args"]
        assert result_model_args == expected_model_args

        result_other = {
            key: result[key] for key in result if key not in {"model_args", "data"}
        }

        expected_other = {
            key: expected[key] for key in expected if key not in {"model_args", "data"}
        }
        assert result_other == expected_other

    @staticmethod
    def test_input_raises_w_data_and_ucm_model(causal_impact, ucm_model):
        # Test inconsistent input (must not provide both data and ucm_model)
        with pytest.raises(SyntaxError) as excinfo:
            causal_impact._format_input(
                causal_impact.params["data"],
                causal_impact.params["pre_period"],
                causal_impact.params["post_period"],
                causal_impact.params["model_args"],
                ucm_model,
                [1, 2, 3],
                causal_impact.params["alpha"],
            )
        assert str(excinfo.value) == (
            "Must either provide ``data``, "
            "``pre_period`` ,``post_period``, ``model_args`` or "
            "``ucm_modeland ``post_period_response``"
        )

    @staticmethod
    def test_input_w_ucm_input(ucm_model, impact_ucm):
        expected = {
            "data": None,
            "pre_period": None,
            "post_period": None,
            "ucm_model": impact_ucm.params["ucm_model"],
            "post_period_response": impact_ucm.params["post_period_response"],
            "alpha": impact_ucm.params["alpha"],
        }

        result = impact_ucm._format_input(
            None,
            None,
            None,
            None,
            impact_ucm.params["ucm_model"],
            impact_ucm.params["post_period_response"],
            impact_ucm.params["alpha"],
        )
        result.pop("model_args")
        assert result == expected

    @staticmethod
    def test_format_output_is_df(causal_impact):
        # Test that <data> is converted to pandas DataFrame
        expected_data = pd.DataFrame(np.arange(0, 8).reshape(4, 2), index=[0, 1, 2, 3])

        funny_datas = [
            pd.DataFrame([[0, 1], [2, 3], [4, 5], [6, 7]]),
            pd.DataFrame(data=[[0, 1], [2, 3], [4, 5], [6, 7]], index=[0, 1, 2, 3]),
            [[0, 1], [2, 3], [4, 5], [6, 7]],
            np.array([[0, 1], [2, 3], [4, 5], [6, 7]]),
        ]

        for funny_data in funny_datas:
            result = causal_impact._format_input(
                funny_data, [0, 2], [3, 3], {}, None, None, 0.05
            )
            assert_array_equal(result["data"].values, expected_data.values)
            assert isinstance(result["data"], pd.DataFrame)

    @staticmethod
    def test_input_w_bad_data(causal_impact):
        text_data = "foo"
        with pytest.raises(ValueError) as excinfo:
            causal_impact._format_input(text_data, [0, 3], [3, 3], {}, None, None, 0.05)
        assert str(excinfo.value) == (
            "could not convert input data to " "Pandas DataFrame"
        )

    @staticmethod
    def test_input_w_bad_pre_period(data, causal_impact):
        bad_pre_periods = [
            1,
            [],
            [1, 2, 3],
            [np.nan, 2],
            [pd.to_datetime(date) for date in ["2011-01-01", "2011-12-31"]],
        ]

        errors_list = [
            "pre_period and post_period must both be lists",
            "pre_period and post_period must both be of length 2",
            "pre_period and post_period must both be of length 2",
            "pre_period and post period must not contain null values",
            (
                "pre_period (object) and post_period (int64) should have the same"
                " class as the time points in the data (int64)"
            ),
        ]

        for idx, bad_pre_period in enumerate(bad_pre_periods):
            with pytest.raises(ValueError) as excinfo:
                causal_impact._format_input(
                    data, bad_pre_period, [1, 2], None, None, None, 0.05
                )
            assert str(excinfo.value) == errors_list[idx]

    @staticmethod
    def test_input_w_bad_post_period(data, causal_impact):
        bad_post_periods = [
            1,
            [],
            [1, 2, 3],
            [np.nan, 2],
            [pd.to_datetime(date) for date in ["2011-01-01", "2011-12-31"]],
        ]

        errors_list = [
            "pre_period and post_period must both be lists",
            "pre_period and post_period must both be of length 2",
            "pre_period and post_period must both be of length 2",
            "pre_period and post period must not contain null values",
            (
                "pre_period (int64) and post_period (object) should have the same"
                " class as the time points in the data (int64)"
            ),
        ]

        for idx, bad_post_period in enumerate(bad_post_periods):
            with pytest.raises(ValueError) as excinfo:
                causal_impact._format_input(
                    data, [1, 2], bad_post_period, None, None, None, 0.05
                )
            assert str(excinfo.value) == errors_list[idx]

    @staticmethod
    def test_input_w_pre_and_post_periods_having_distinct_classes(causal_impact):
        # Test what happens when pre_period/post_period has a different class
        # than the timestamps in <data>
        bad_data = pd.DataFrame(
            data=[1, 2, 3, 4],
            index=["2014-01-01", "2014-01-02", "2014-01-03", "2014-01-04"],
        )

        bad_pre_period = [0.0, 3.0]  # float
        bad_post_period = [3, 3]

        with pytest.raises(ValueError) as excinfo:
            causal_impact._format_input(
                bad_data, bad_pre_period, bad_post_period, None, None, None, 0.05
            )
        assert str(excinfo.value) == (
            "pre_period (float64) and post_period ("
            "int64) should have the same class as the time points in the data"
            " (object)"
        )

        bad_pre_period = [0, 2]  # integer
        bad_post_period = [3, 3]
        with pytest.raises(ValueError) as excinfo:
            causal_impact._format_input(
                bad_data, bad_pre_period, bad_post_period, None, None, None, 0.05
            )
        assert str(excinfo.value) == (
            "pre_period (int64) and post_period ("
            "int64) should have the same class as the time points in the data"
            " (object)"
        )

    @staticmethod
    def test_bad_model_args(data, causal_impact):
        with pytest.raises(TypeError) as excinfo:
            causal_impact._format_input(data, [0, 3], [3, 10], 1000, None, None, 0.05)
        assert str(excinfo.value) == "argument of type 'int' is not iterable"
        with pytest.raises(TypeError) as excinfo:
            causal_impact._format_input(
                data, [0, 3], [3, 10], "ninter=1000", None, None, 0.05
            )
        assert str(excinfo.value) == ("'str' object does not support item assignment")

    @staticmethod
    def test_bad_standardize(data, causal_impact):
        bad_standardize_data = [np.nan, 123, "foo", [True, False]]
        for bad_standardize in bad_standardize_data:
            bad_model_args = {"standardize_data": bad_standardize}
            with pytest.raises(ValueError) as excinfo:
                causal_impact._format_input(
                    data, [0, 3], [3, 10], bad_model_args, None, None, 0.05
                )
            assert str(excinfo.value) == (
                "model_args.standardize_data must be a boolean value"
            )

    @staticmethod
    def test_bad_post_period_response(causal_impact, ucm_model):
        with pytest.raises(ValueError) as excinfo:
            causal_impact._format_input(
                None, None, None, None, ucm_model, pd.to_datetime("2011-01-01"), 0.05
            )
        assert str(excinfo.value) == ("post_period_response must be list-like")

        with pytest.raises(ValueError) as excinfo:
            causal_impact._format_input(None, None, None, None, ucm_model, True, 0.05)
        assert str(excinfo.value) == ("post_period_response must be list-like")

        with pytest.raises(ValueError) as excinfo:
            causal_impact._format_input(
                None, None, None, None, ucm_model, [pd.to_datetime("2018-01-01")], 0.05
            )
        assert str(excinfo.value) == (
            "post_period_response should not be datetime values"
        )

        with pytest.raises(ValueError) as excinfo:
            causal_impact._format_input(None, None, None, None, ucm_model, [2j], 0.05)
        assert str(excinfo.value) == (
            "post_period_response must contain all real values"
        )

    @staticmethod
    def test_bad_alpha(data, causal_impact):
        bad_alphas = [None, np.nan, -1, 0, 1, [0.8, 0.9], "0.1"]
        for bad_alpha in bad_alphas:
            with pytest.raises(ValueError) as excinfo:
                causal_impact._format_input(
                    data, [0, 3], [3, 10], {}, None, None, bad_alpha
                )
        assert str(excinfo.value) == "alpha must be a real number"

    @staticmethod
    def test_bad_ucm(data, causal_impact):
        from statsmodels.tsa.statespace.sarimax import SARIMAX

        bad_ucm = SARIMAX(endog=[1, 2, 3, 4])

        with pytest.raises(ValueError) as excinfo:
            causal_impact._format_input(None, None, None, None, bad_ucm, [2], 0.05)
        assert (
            str(excinfo.value)
            == "ucm_model must be an object of class "
            + "statsmodels.tsa.statespace.structural.UnobservedComponents "
            + "instead received statsmodels.tsa.statespace.sarimax.SARIMAX"
        )

    @staticmethod
    def test_input_w_date_column():
        data = pd.DataFrame(np.random.randn(100, 2), columns=["x1", "x2"])
        data["date"] = pd.date_range(start="2018-01-01", periods=100)
        data = data[["date", "x1", "x2"]]
        pre_period = ["2018-01-01", "2018-02-10"]
        post_period = ["2018-02-11", "2018-4-10"]
        causal_impact = CausalImpact(data, pre_period, post_period, {})
        data = data.set_index("date")
        pre_period = [pd.to_datetime(e) for e in pre_period]
        post_period = [pd.to_datetime(e) for e in post_period]

        expected = {
            "data": data,
            "pre_period": pre_period,
            "post_period": post_period,
            "model_args": causal_impact.params["model_args"],
            "ucm_model": None,
            "post_period_response": None,
            "alpha": causal_impact.params["alpha"],
        }
        result = causal_impact._format_input(
            causal_impact.params["data"],
            causal_impact.params["pre_period"],
            causal_impact.params["post_period"],
            causal_impact.params["model_args"],
            None,
            None,
            causal_impact.params["alpha"],
        )

        result_data = result["data"]
        expected_data = expected["data"]
        assert_frame_equal(result_data, expected_data)

        result_model_args = result["model_args"]
        expected_model_args = expected["model_args"]
        assert result_model_args == expected_model_args

        result_other = {
            key: result[key] for key in result if key not in {"model_args", "data"}
        }

        expected_other = {
            key: expected[key] for key in expected if key not in {"model_args", "data"}
        }
        assert result_other == expected_other

    @staticmethod
    def test_input_w_time_column():
        data = pd.DataFrame(np.random.randn(100, 2), columns=["x1", "x2"])
        data["time"] = pd.date_range(start="2018-01-01", periods=100)
        data = data[["time", "x1", "x2"]]
        pre_period = ["2018-01-01", "2018-02-10"]
        post_period = ["2018-02-11", "2018-4-10"]

        causal_impact = CausalImpact(data, pre_period, post_period, {})

        data = data.set_index("time")
        pre_period = [pd.to_datetime(e) for e in pre_period]
        post_period = [pd.to_datetime(e) for e in post_period]

        expected = {
            "data": data,
            "pre_period": pre_period,
            "post_period": post_period,
            "model_args": causal_impact.params["model_args"],
            "ucm_model": None,
            "post_period_response": None,
            "alpha": causal_impact.params["alpha"],
        }
        result = causal_impact._format_input(
            causal_impact.params["data"],
            causal_impact.params["pre_period"],
            causal_impact.params["post_period"],
            causal_impact.params["model_args"],
            None,
            None,
            causal_impact.params["alpha"],
        )

        result_data = result["data"]
        expected_data = expected["data"]
        assert_frame_equal(result_data, expected_data)

        result_model_args = result["model_args"]
        expected_model_args = expected["model_args"]
        assert result_model_args == expected_model_args

        result_other = {
            key: result[key] for key in result if key not in {"model_args", "data"}
        }

        expected_other = {
            key: expected[key] for key in expected if key not in {"model_args", "data"}
        }
        assert result_other == expected_other

    @staticmethod
    def test_input_w_just_2_points_raises_exception():
        data = pd.DataFrame(np.random.randn(2, 2), columns=["x1", "x2"])
        causal_impact = CausalImpact(data, [0, 0], [1, 1], {})

        with pytest.raises(ValueError) as excinfo:
            causal_impact._format_input(
                causal_impact.params["data"],
                causal_impact.params["pre_period"],
                causal_impact.params["post_period"],
                causal_impact.params["model_args"],
                None,
                None,
                causal_impact.params["alpha"],
            )
        assert str(excinfo.value) == "data must have at least 3 time points"

    @staticmethod
    def test_input_covariates_w_nan_value_raises():
        data = np.array([[1, 1, 2], [1, 2, 3], [1, 3, 4], [1, np.nan, 5], [1, 6, 7]])
        data = pd.DataFrame(data, columns=["y", "x1", "x2"])
        causal_impact = CausalImpact(data, [0, 3], [3, 4], {})

        with pytest.raises(ValueError) as excinfo:
            causal_impact._format_input(
                causal_impact.params["data"],
                causal_impact.params["pre_period"],
                causal_impact.params["post_period"],
                causal_impact.params["model_args"],
                None,
                None,
                causal_impact.params["alpha"],
            )
        assert str(excinfo.value) == "covariates must not contain null values"

    @staticmethod
    def test_int_index_pre_period_contains_float(causal_impact, pre_period):
        expected = {
            "data": causal_impact.params["data"],
            "pre_period": causal_impact.params["pre_period"],
            "post_period": causal_impact.params["post_period"],
            "model_args": causal_impact.params["model_args"],
            "ucm_model": None,
            "post_period_response": None,
            "alpha": causal_impact.params["alpha"],
        }
        result = causal_impact._format_input(
            causal_impact.params["data"],
            [float(pre_period[0]), pre_period[1]],
            causal_impact.params["post_period"],
            causal_impact.params["model_args"],
            None,
            None,
            causal_impact.params["alpha"],
        )

        result_data = result["data"]
        expected_data = expected["data"]
        assert_frame_equal(result_data, expected_data)

        result_model_args = result["model_args"]
        expected_model_args = expected["model_args"]
        assert result_model_args == expected_model_args

        result_other = {
            key: result[key] for key in result if key not in {"model_args", "data"}
        }

        expected_other = {
            key: expected[key] for key in expected if key not in {"model_args", "data"}
        }
        assert result_other == expected_other

    @staticmethod
    def test_float_index_pre_period_contains_int():
        data = np.random.randn(200, 3)
        data = pd.DataFrame(data, columns=["y", "x1", "x2"])
        data = data.set_index(np.array([float(i) for i in range(200)]))
        causal_impact = CausalImpact(data, [0, 3], [3, 4], {})

        expected = {
            "data": causal_impact.params["data"],
            "pre_period": causal_impact.params["pre_period"],
            "post_period": causal_impact.params["post_period"],
            "model_args": causal_impact.params["model_args"],
            "ucm_model": None,
            "post_period_response": None,
            "alpha": causal_impact.params["alpha"],
        }
        result = causal_impact._format_input(
            causal_impact.params["data"],
            causal_impact.params["pre_period"],
            causal_impact.params["post_period"],
            causal_impact.params["model_args"],
            None,
            None,
            causal_impact.params["alpha"],
        )

        result_data = result["data"]
        expected_data = expected["data"]
        assert_frame_equal(result_data, expected_data)

        result_model_args = result["model_args"]
        expected_model_args = expected["model_args"]
        assert result_model_args == expected_model_args

        result_other = {
            key: result[key] for key in result if key not in {"model_args", "data"}
        }
        expected_other = {
            key: expected[key] for key in expected if key not in {"model_args", "data"}
        }
        assert result_other == expected_other

    @staticmethod
    def test_pre_period_in_conflict_w_post_period():
        data = pd.DataFrame(np.random.randn(20, 2), columns=["x1", "x2"])
        causal_impact = CausalImpact(data, [0, 10], [9, 20], {})

        with pytest.raises(ValueError) as excinfo:
            causal_impact._format_input(
                causal_impact.params["data"],
                causal_impact.params["pre_period"],
                causal_impact.params["post_period"],
                causal_impact.params["model_args"],
                None,
                None,
                causal_impact.params["alpha"],
            )
        assert str(excinfo.value) == (
            "post period must start at least 1 observation after the end of "
            "the pre_period"
        )

        causal_impact = CausalImpact(data, [0, 10], [11, 9], {})
        with pytest.raises(ValueError) as excinfo:
            causal_impact._format_input(
                causal_impact.params["data"],
                causal_impact.params["pre_period"],
                causal_impact.params["post_period"],
                causal_impact.params["model_args"],
                None,
                None,
                causal_impact.params["alpha"],
            )
        assert str(excinfo.value) == (
            "post_period[1] must not be earlier than post_period[0]"
        )

        causal_impact = CausalImpact(data, [0, 10], [11, 9], {})
        with pytest.raises(ValueError) as excinfo:
            causal_impact._format_input(
                causal_impact.params["data"],
                causal_impact.params["pre_period"],
                causal_impact.params["post_period"],
                causal_impact.params["model_args"],
                None,
                None,
                causal_impact.params["alpha"],
            )
        assert str(excinfo.value) == (
            "post_period[1] must not be earlier than post_period[0]"
        )


class TestRunWithData:
    @staticmethod
    def test_missing_input():
        with pytest.raises(SyntaxError):
            impact = CausalImpact()
            impact.run()

    @staticmethod
    def test_unlabelled_pandas_series(expected_columns, pre_period, post_period):
        model_args = {"niter": 123, "standardize_data": False}
        alpha = 0.05
        data = pd.DataFrame(np.random.randn(200, 3))
        causal_impact = CausalImpact(
            data.values, pre_period, post_period, model_args, None, None, alpha, "MLE"
        )

        causal_impact.run()
        actual_columns = list(causal_impact.inferences.columns)
        assert actual_columns == expected_columns

    @staticmethod
    def test_other_formats(expected_columns, pre_period, post_period):
        # Test other data formats
        model_args = {"niter": 100, "standardize_data": True}

        # labelled dataframe
        data = pd.DataFrame(np.random.randn(200, 3), columns=["a", "b", "c"])
        impact = CausalImpact(data, pre_period, post_period, model_args)
        impact.run()
        actual_columns = list(impact.inferences.columns)
        assert actual_columns == expected_columns

        # numpy array
        data = np.random.randn(200, 3)
        impact = CausalImpact(data, pre_period, post_period, model_args)
        impact.run()
        actual_columns = list(impact.inferences.columns)
        assert actual_columns == expected_columns

        # list of lists
        data = np.random.randn(200, 2).tolist()
        impact = CausalImpact(data, pre_period, post_period, model_args)
        impact.run()
        actual_columns = list(impact.inferences.columns)
        assert actual_columns == expected_columns

    @staticmethod
    def test_frame_w_no_exog(pre_period, post_period):
        data = np.random.randn(200)
        impact = CausalImpact(data, pre_period, post_period, {})
        with pytest.raises(ValueError) as excinfo:
            impact.run()
        assert str(excinfo.value) == "data contains no exogenous variables"

    @staticmethod
    def test_missing_pre_period_data(data, pre_period, post_period):
        model_data = data.copy()
        model_data.iloc[3:5, 0] = np.nan
        impact = CausalImpact(model_data, pre_period, post_period)
        impact.run()
        assert len(impact.inferences) == len(model_data)

    @staticmethod
    def test_pre_period_starts_after_beginning_of_data(data):
        pre_period = [3, 100]
        impact = CausalImpact(data, pre_period, [101, 199])
        impact.run()
        np.testing.assert_array_equal(impact.inferences.response.values, data.y.values)
        assert np.all(pd.isnull(impact.inferences.iloc[0 : pre_period[0], 2:]))

    @staticmethod
    def test_post_period_finishes_before_end_of_data(data, pre_period):
        post_period = [101, 197]
        impact = CausalImpact(data, pre_period, post_period)
        impact.run()
        np.testing.assert_array_equal(impact.inferences.response.values, data.y.values)
        assert np.all(pd.isnull(impact.inferences.iloc[-2:, 2:]))

    @staticmethod
    def test_gap_between_pre_and_post_periods(data, pre_period):
        post_period = [120, 199]
        impact = CausalImpact(data, pre_period, post_period)
        impact.run()
        assert np.all(
            pd.isnull(impact.inferences.loc[101:119, impact.inferences.columns[2:]])
        )

    @staticmethod
    def test_late_start_early_finish_and_gap_between_periods(data):
        pre_period = [3, 80]
        post_period = [120, 197]
        impact = CausalImpact(data, pre_period, post_period)
        impact.run()
        assert np.all(
            pd.isnull(impact.inferences.loc[:2, impact.inferences.columns[2:]])
        )
        assert np.all(
            pd.isnull(impact.inferences.loc[81:119, impact.inferences.columns[2:]])
        )
        assert np.all(
            pd.isnull(impact.inferences.loc[198:, impact.inferences.columns[2:]])
        )

    @staticmethod
    def test_pre_period_lower_than_data_index_min(data):
        pre_period = [-1, 100]
        post_period = [101, 199]
        impact = CausalImpact(data, pre_period, post_period)
        impact.run()
        assert impact.params["pre_period"] == [0, 100]

    @staticmethod
    def test_post_period_bigger_than_data_index_max(data):
        pre_period = [0, 100]
        post_period = [101, 300]
        impact = CausalImpact(data, pre_period, post_period)
        impact.run()
        assert impact.params["post_period"] == [101, 199]

    @staticmethod
    def test_missing_values_in_pre_period_y(pre_period, post_period):
        """Test that all columns in the result series except the
        point predictions have missing values at the time points the
        result time series has missing values."""
        data = pd.DataFrame(np.random.randn(200, 3), columns=["y", "x1", "x2"])
        data.iloc[95:100, 0] = np.nan

        impact = CausalImpact(data, pre_period, post_period)
        impact.run()

        predicted_cols = [
            impact.inferences.columns.get_loc(col)
            for col in impact.inferences.columns
            if ("response" not in col and "point_effect" not in col)
        ]

        effect_cols = [
            impact.inferences.columns.get_loc(col)
            for col in impact.inferences.columns
            if "point_effect" in col
        ]

        response_cols = [
            impact.inferences.columns.get_loc(col)
            for col in impact.inferences.columns
            if "response" in col
        ]

        assert np.all(np.isnan(impact.inferences.iloc[95:100, response_cols]))
        assert np.all(np.isnan(impact.inferences.iloc[95:100, effect_cols]))
        TestCase().assertFalse(
            np.any(np.isnan(impact.inferences.iloc[95:100, predicted_cols]))
        )
        TestCase().assertFalse(np.any(np.isnan(impact.inferences.iloc[:95, :])))
        TestCase().assertFalse(np.any(np.isnan(impact.inferences.iloc[101:, :])))


class TestRunWithUCM:
    @staticmethod
    def test_regular_run(expected_columns, impact_ucm):
        impact_ucm.run()
        actual_columns = list(impact_ucm.inferences.columns)
        assert actual_columns == expected_columns


class TestSummary:
    @staticmethod
    def test_summary(inference_input):
        inferences_df = pd.DataFrame(inference_input)
        causal = CausalImpact()

        params = {"alpha": 0.05, "post_period": [2, 4]}

        causal.params = params
        causal.inferences = inferences_df

        expected = [
            [3, 7],
            [3, 7],
            [[3, 3], [7, 7]],
            [" ", " "],
            [0, 0],
            [[0, 0], [0, 0]],
            [" ", " "],
            ["-2.8%", "-2.8%"],
            [["0.0%", "-11.1%"], ["0.0%", "-11.1%"]],
            [" ", " "],
            ["0.0%", " "],
            ["100.0%", " "],
        ]

        expected = pd.DataFrame(
            expected,
            columns=["Average", "Cumulative"],
            index=[
                "Actual",
                "Predicted",
                "95% CI",
                " ",
                "Absolute Effect",
                "95% CI",
                " ",
                "Relative Effect",
                "95% CI",
                " ",
                "P-value",
                "Prob. of Causal Effect",
            ],
        )

        tmpdir = mkdtemp()
        tmp_expected = "tmp_expected"
        tmp_result = "tmp_test_summary"

        result_file = os.path.join(tmpdir, tmp_result)
        expected_file = os.path.join(tmpdir, tmp_expected)

        expected.to_csv(expected_file)
        expected_str = open(expected_file).read()

        causal.summary(path=result_file)

        result = open(result_file).read()
        assert result == expected_str

    @staticmethod
    def test_summary_w_report_output(
        monkeypatch, inference_input, summary_report_filename
    ):
        inferences_df = pd.DataFrame(inference_input)
        causal = CausalImpact()

        params = {"alpha": 0.05, "post_period": [2, 4]}

        causal.params = params
        causal.inferences = inferences_df

        dedent_mock = mock.Mock()

        expected = open(summary_report_filename).read()
        expected = re.sub(r"\s+", " ", expected)
        expected = expected.strip()

        tmpdir = mkdtemp()
        tmp_file = os.path.join(tmpdir, "summary_test")

        def dedent_side_effect(msg):
            with open(tmp_file, "a") as file_obj:
                msg = re.sub(r"\s+", " ", msg)
                msg = msg.strip()
                file_obj.write(msg)
            return msg

        dedent_mock.side_effect = dedent_side_effect
        monkeypatch.setattr("textwrap.dedent", dedent_mock)

        causal.summary(output="report")
        result_str = open(tmp_file, "r").read()
        assert result_str == expected

    @staticmethod
    def test_summary_wrong_argument_raises(inference_input):
        inferences_df = pd.DataFrame(inference_input)
        causal = CausalImpact()

        params = {"alpha": 0.05, "post_period": [2, 4]}

        causal.params = params
        causal.inferences = inferences_df

        with pytest.raises(ValueError):
            causal.summary(output="wrong_argument")


class TestPlot:
    # @patch('causalimpact.data')
    @staticmethod
    def test_plot(monkeypatch):

        params = {"alpha": 0.05, "post_period": [2, 4], "pre_period": [0, 1]}
        inferences_mock = mock.MagicMock()

        class EnhancedDict:
            @property
            def index(self):
                return [0, 1]

            @property
            def response(self):
                return "y obs"

            @property
            def point_pred(self):
                return "points predicted"

            @property
            def point_pred_lower(self):
                return "lower predictions"

            @property
            def point_pred_upper(self):
                return "upper predictions"

            @property
            def point_effect(self):
                return "lift"

            @property
            def point_effect_lower(self):
                return "point effect lower"

            @property
            def point_effect_upper(self):
                return "point effect upper"

            @property
            def cum_effect(self):
                return "cum effect"

            @property
            def cum_effect_lower(self):
                return "cum effect lower"

            @property
            def cum_effect_upper(self):
                return "cum effect upper"

        def getitem(name):
            return EnhancedDict()

        inferences_mock.iloc.__getitem__.side_effect = getitem

        class Data:
            @property
            def index(self):
                return "index"

            @property
            def shape(self):
                return [(1, 2)]

        data_mock = Data()

        plot_mock = mock.Mock()
        np_zeros_mock = mock.Mock()
        np_zeros_mock.side_effect = lambda x: [0, 0]

        get_lib_mock = mock.Mock(return_value=plot_mock)
        monkeypatch.setattr("causalimpact.analysis.get_matplotlib", get_lib_mock)

        monkeypatch.setattr("numpy.zeros", np_zeros_mock)

        causal = CausalImpact()
        causal.params = params
        causal.inferences = inferences_mock
        causal.data = data_mock

        causal.plot(panels=["original", "pointwise", "cumulative"])
        causal.plot(panels=["pointwise", "cumulative"])

        causal.plot(panels=["original"])
        plot_mock.plot.assert_any_call("y obs", "k", label="endog", linewidth=2)
        plot_mock.plot.assert_any_call(
            "points predicted", "r--", label="model", linewidth=2
        )

        plot_mock.fill_between.assert_any_call(
            [0, 1],
            "lower predictions",
            "upper predictions",
            facecolor="gray",
            interpolate=True,
            alpha=0.25,
        )

        causal.plot(panels=["pointwise"])

        plot_mock.plot.assert_any_call("lift", "r--", linewidth=2)
        plot_mock.plot.assert_any_call("index", [0, 0], "g-", linewidth=2)

        causal.plot(panels=["cumulative"])

        plot_mock.plot.assert_any_call([0, 1], "cum effect", "r--", linewidth=2)
        plot_mock.plot.assert_any_call("index", [0, 0], "g-", linewidth=2)



"""Constructs and fits the statespace model.

Contains the construct_model and model_fit functions that are called in analysis.py.
"""

import numpy as np
import pandas as pd
import pymc as pm
import pytensor.tensor as at
from pytensor.graph.op import Op


def observations_ill_conditioned(y):
    """Checks whether the response variable (i.e., the series of observations
    for the dependent variable y) are ill-conditioned. For example, the series
    might contain too few non-NA values. In such cases, inference will be
    aborted.

    Args:
        y: observed series (Pandas Series)

    Returns:
        True if something is wrong with the observations; False otherwise.
    """

    if y is None:
        raise ValueError("y cannot be None")
    if not (len(y) > 1):
        raise ValueError("y must have len > 1")

    # All NA?
    if np.all(pd.isnull(y)):
        raise ValueError("Aborting inference due to input series being all " "null.")
    elif len(y[pd.notnull(y)]) < 3:
        # Fewer than 3 non-NA values?
        raise ValueError(
            "Aborting inference due to fewer than 3 nonnull " "values in input."
        )
    # Constant series?
    elif y.std(skipna=True) == 0:
        raise ValueError("Aborting inference due to input series being " "constant")
    return False


def construct_model(data, model_args=None):
    """Specifies the model and performs inference. Inference means using a
    technique that combines Kalman Filters with Maximum Likelihood Estimators
    methods to fit the parameters that best explain the observed data.

    Args:
      data: time series of response variable and optional covariates
      model_args: optional list of additional model arguments

    Returns:
      An Unobserved Components Model, as returned by UnobservedComponents()
    """
    if model_args is None:
        model_args = {}
    from statsmodels.tsa.statespace.structural import UnobservedComponents

    y = data.iloc[:, 0]

    observations_ill_conditioned(y)

    # LocalLevel specification of statespace
    ss = {"endog": y.values, "level": "llevel"}

    # No regression?
    if len(data.columns) > 1:
        # Static regression
        if not model_args.get("dynamic_regression"):
            ss["exog"] = data.iloc[:, 1:].values
        # Dynamic regression
        else:
            raise NotImplementedError()
    mod = UnobservedComponents(**ss)
    return mod


class Loglike(Op):
    """Theano LogLike wrapper that allow  PyMC3 to compute the likelihood
    and Jacobian in a way that it can make use of."""

    itypes = [at.dvector]  # expects a vector of parameter values when called
    otypes = [at.dscalar]  # outputs a single scalar value (the log likelihood)

    def __init__(self, model):
        self.model = model
        self.score = Score(self.model)

    def perform(self, node, inputs, outputs):
        (theta,) = inputs  # contains the vector of parameters
        llf = self.model.loglike(theta)
        outputs[0][0] = np.array(llf)  # output the log-likelihood

    def grad(self, inputs, g):
        # the method that calculates the gradients - it actually returns the
        # vector-Jacobian product - g[0] is a vector of parameter values
        (theta,) = inputs  # our parameters
        out = [g[0] * self.score(theta)]
        return out


class Score(Op):
    """Theano Score wrapper that allow  PyMC3 to compute the likelihood and
    Jacobian in a way that it can make use of."""

    itypes = [at.dvector]
    otypes = [at.dvector]

    def __init__(self, model):
        self.model = model

    def perform(self, node, inputs, outputs):
        (theta,) = inputs
        outputs[0][0] = self.model.score(theta)


class ModelResults:
    """ModelResults class containing everything needed for inference
    intended to allow extension to other models (e.g. tensorflow)

    Parameters
    ----------
    ucm_model : statsmodels.tsa.statespace.structural.UnobservedComponents
        The constructed UCM model being fit
    results :
    estimation : string
        The estimation method. Options are "MLE" or "pymc"

    """

    def __init__(self, ucm_model, results, estimation) -> None:
        self.results = results
        self.estimation = estimation
        self.model_nobs = ucm_model.nobs

    def get_prediction(self, start=None, end=None):
        """
        In-sample prediction and out-of-sample forecasting


        Parameters
        ----------
        start : int
            Zero-indexed observation number at which to start forecasting,
            i.e., the first forecast is start. Can also be a date string to
            parse or a datetime type. Default is the the zeroth observation.
        end : int
            Zero-indexed observation number at which to end forecasting,
            i.e., the last forecast is end. Can also be a date string to
            parse or a datetime type. However, if the dates index does not
            have a fixed frequency, end must be an integer index if you want
            out of sample prediction. Default is the last observation in the
            sample.

        Returns
        -------
        ModelPredictions
        """
        predictions = self.results.get_prediction(start=start, end=end)
        return predictions

    def get_forecast(self, df_post, alpha):
        forecast = self.results.get_forecast(
            steps=len(df_post), exog=df_post.iloc[:, 1:], alpha=alpha
        )
        return forecast

    def summary(self):
        return self.results.summary()


def model_fit(model, estimation, model_args):
    """Fits the model and returns a ModelResults object.

    Uses the chosen estimation option to fit the model and
    return a ModelResults object that is agnostic of
    estimation approach.

    Parameters:
    -----------
    model : statsmodels.tsa.statespace.structural.UnobservedComponents
    estimation : str
        Either 'MLE' or 'pymc'.
    model_args : dict
        possible args for MLE are:
        niter: int
        possible args for pymc are:
        ndraws: int
            number of draws from the distribution
        nburn: int
            number of "burn-in points" (which will be discarded)

    """
    if estimation == "MLE":
        trained_model = model.fit(maxiter=model_args["niter"])
        model_results = ModelResults(model, trained_model, estimation)
        return model_results
    elif estimation == "pymc":
        loglike = Loglike(model)
        with pm.Model():
            # Priors
            sigma2irregular = pm.InverseGamma("sigma2.irregular", 1, 1)
            sigma2level = pm.InverseGamma("sigma2.level", 1, 1)
            if model.exog is None:
                # convert variables to tensor vectors
                theta = at.as_tensor_variable([sigma2irregular, sigma2level])
            else:
                # prior for regressors
                betax1 = pm.Laplace("beta.x1", mu=0, b=1.0 / 0.7)
                # convert variables to tensor vectors
                theta = at.as_tensor_variable([sigma2irregular, sigma2level, betax1])
            # use a DensityDist (use a lambda function to "call" the Op)
            pm.Potential("likelihood", loglike(theta))

            # Draw samples
            trace = pm.sample(
                model_args["ndraws"],
                tune=model_args["nburn"],
                return_inferencedata=True,
                cores=4,
                compute_convergence_checks=False,
            )
        # Retrieve the posterior means
        params = pm.summary(trace)["mean"].values

        # Construct results using these posterior means as parameter values
        results = model.smooth(params)
        model_results = ModelResults(model, results, estimation)
        return model_results



import pandas as pd


def standardize_all_variables(data, pre_period, post_period):
    """Standardize all columns of a given time series.
    Args:
        data: Pandas DataFrame with one or more columns

    Returns:
        dict:
            data: standardized data
            UnStandardize: function for undoing the transformation of the
            first column in the provided data
    """
    if not isinstance(data, pd.DataFrame):
        raise ValueError("``data`` must be of type `pandas.DataFrame`")

    if not (
        pd.api.types.is_list_like(pre_period) and pd.api.types.is_list_like(post_period)
    ):
        raise ValueError("``pre_period`` and ``post_period``must be listlike")

    data_mu = data.loc[pre_period[0] : pre_period[1], :].mean(skipna=True)
    data_sd = data.loc[pre_period[0] : pre_period[1], :].std(skipna=True, ddof=0)
    data = data - data_mu
    data_sd = data_sd.fillna(1)

    data[data != 0] = data[data != 0] / data_sd
    y_mu = data_mu[0]
    y_sd = data_sd[0]

    data_pre = data.loc[pre_period[0] : pre_period[1], :]
    data_post = data.loc[post_period[0] : post_period[1], :]

    return {
        "data_pre": data_pre,
        "data_post": data_post,
        "orig_std_params": (y_mu, y_sd),
    }


def unstandardize(data, orig_std_params):
    """Function for reversing the standardization of the first column in the
    provided data.
    """
    data = pd.DataFrame(data)
    y_mu = orig_std_params[0]
    y_sd = orig_std_params[1]
    data = data.mul(y_sd, axis=1)
    data = data.add(y_mu, axis=1)
    return data


def df_print(data, path=None):
    if path:
        data.to_csv(path)
    print(data)


def get_matplotlib():  # pragma: no cover
    """Wrapper function to facilitate unit testing the `plot` tool by removing
    the strong dependencies of matplotlib.

    Returns:
        module matplotlib.pyplot
    """
    import matplotlib.pyplot as plt

    return plt
